from django.db import models
from django.contrib.auth.models import User

class Revision(models.Model):
    revision_type = 'base'

    url = models.CharField(max_length=256)
    created_date = models.DateTimeField(auto_now_add=True, null=False, blank=False)
    created_by = models.ForeignKey(User, null=True, blank=True)
    parent = models.ForeignKey('self', null=True, blank=True)

    @property
    def revision_id(self):
        return self.id

    def clone(self):
        return Revision(url=self.url, parent=self)

    def descended_from(self, other_rev):
        if self.parent is None:
            return False
        elif self.parent.revision_id == other_rev.revision_id:
            return True
        else:
            return self.parent.descended_from(other_rev)

    def history(self, stop_at=None):
        revisions = list()
        if not self.revision_id:
            return revisions
        rev = self
        while rev is not None:
            revisions.append(rev)
            if rev.revision_id == stop_at:
                break
            rev = rev.parent
        return revisions

    def __unicode__(self):
        if self.parent is not None:
            return '%s [rev=%s] (parent: rev=%s)' % (self.url, self.revision_id, self.parent.revision_id)
        else:
            return '%s [rev=%s]' % (self.url, self.revision_id)


class ContentRevision(Revision):
    revision_type = 'content'

    content = models.TextField()

    def clone(self):
        return ContentRevision(url=self.url, content=self.content, parent=self)


class RedirectRevision(Revision):
    revision_type = 'redirect'

    redirect_url = models.CharField(max_length=256)

    def clone(self):
        return RedirectRevision(
            url=self.url,
            redirect_url=self.redirect_url,
            parent=self)

    def __unicode__(self):
        if self.parent is not None:
            return '%s -> %s [rev=%s] (parent: rev=%s)' % (self.url, self.redirect_url, self.revision_id, self.parent.revision_id)
        else:
            return '%s -> %s [rev=%s]' % (self.url, self.redirect_url, self.revision_id)


class PanicException(Exception):
    pass


class NoPageForUrl(Exception):

    def __init__(self, view, url):
        super(NoPageForUrl, self).__init__(
            'Could not find a page for url: %s' % url)
        self.view = view
        self.url = url


class SiteView(models.Model):

    name = models.CharField(max_length=64)
    read_only = models.BooleanField(null=False, blank=False, default=False)
    obsolete = models.BooleanField(null=False, blank=False, default=False)
    parent = models.ForeignKey('self', null=True, blank=True)
    pages = models.ManyToManyField(Revision, related_name='views', null=True, blank=True)

    @property
    def view_id(self):
        return self.id

    def get_page(self, url):
        """ return a page for this view based on its URL"""
        try:
            base_page = self.pages.get(url=url)
        except Revision.DoesNotExist:
            raise NoPageForUrl(self, url)
        except:
            raise PanicException(
                "An error occured lookup up this url: %s" % url)

        # Attempt to cast as a child model
        try:
            return base_page.contentrevision
        except ContentRevision.DoesNotExist:
            pass

        try:
            return base_page.redirectrevision
        except RedirectRevision.DoesNotExist:
            pass

        return base_page

    def update_page(self, revision):
        """ Updates the view to use the given revision """
        if self.read_only:
            raise PanicException("Updating a read-only view is not allowed")
        else:
            return self.set_page(revision)
            
    def set_page(self, revision):
        """ sets a given revision to be this view's current revision """
        if revision.parent is not None:
            try:
                current = self.pages.get(id=revision.parent.id)
                self.pages.remove(current)
            except Revision.DoesNotExist:
                # We don't have the parent revision
                pass
        try:
            current = self.pages.get(url=revision.url)
            self.pages.remove(current)
        except Revision.DoesNotExist:
            # We don't have a revision at this url
            pass

        self.pages.add(revision)
        return revision

    def move_page(self, revision, new_url):
        if self.read_only:
            raise PanicException("Updating a read-only view is not allowed")
        else:
            new_rev = revision.clone()
            new_rev.url = new_url
            new_rev.save()
            self.set_page(new_rev)

            redirect = create_revision(
                RedirectRevision,
                url=revision.url,
                redirect_url=new_url)
            self.set_page(redirect)
            return new_rev

    def branch(self, new_name):
        new_view = create_view(view_name=new_name, parent_view=self)
        new_view.pages.add(*list(self.pages.all()))
        return new_view

    def diff(self, other_view):
        additions = dict()
        changes = dict()
        deletions = dict()

        current_pages = dict([(rev.url, self.get_page(rev.url)) for rev in self.pages.all()])
        other_pages = dict([(rev.url, other_view.get_page(rev.url)) for rev in other_view.pages.all()])

        for path in current_pages.keys():
            if not path in other_pages:
                additions[path] = current_pages[path]
            elif other_pages[path].revision_id != current_pages[path].revision_id:
                # If the other view's revision is descended from this view's
                # revision, then it's a change in that view, not this one
                if other_pages[path].descended_from(current_pages[path]):
                    continue

                # If the current revision is a direct descendent of the other
                # view's revision, a history is redundant
                if current_pages[path].parent.revision_id == other_pages[path].revision_id:
                    changes[path] = (other_pages[path], current_pages[path], None)
                else:
                    # TODO: Should find a common anscestor to inform the user
                    # when the two diverged
                    changes[path] = (other_pages[path], current_pages[path], current_pages[path].history(other_pages[path].revision_id))
        deletions = dict([(rev.url, rev) for rev in other_pages.values() if rev.url not in current_pages])
        
            
        return (additions, changes, deletions)

    def update(self, from_view):
        # update pages in this view to match the given view
        (additions, changes, deletions) = from_view.diff(self)
        for page in additions.values():
            self.set_page(page)

        for page in changes.values():
            self.set_page(page[1])

        for page in deletions.values():
            self.pages.remove(page)
        
    def __unicode__(self):
        return self.name


def get_view(view_id):
    """ Take an id and return a reference to the view """
    try:
        return SiteView.objects.get(id=view_id)
    except SiteView.DoesNotExist:
        return None

def get_current_view(request):
    if hasattr(request, 'session') and 'current_view' in request.session:
        try:
            return SiteView.objects.get(pk=request.session['current_view'])
        except:
            # Couldn't locate the session's view, fall through to the default
            # view selection process below
            pass

    try:
        # Get the first view in the database
        view = SiteView.objects.all()[0]
    except IndexError:
        # No views have been defined, so we'll make one
        view = create_view(view_name='Live', read_only=True, parent_view=None)
        
    set_current_view(request, view)
    return view

def set_current_view(request, view):
    request.session['current_view'] = view.view_id
    
def create_view(view_name, parent_view=None, read_only=False, obsolete=False):
    """ create a new view with parent_view as a reference point """
    return SiteView.objects.create(name=view_name, parent=parent_view, read_only=read_only, obsolete=obsolete)


def create_revision(RevisionClass, *args, **kargs):
    """ create a revision of the given class and immediately add it """
    rev = RevisionClass(*args, **kargs)
    return add_revision(rev)


def add_revision(revision):
    """ stores a revision in the system """
    revision.save()
    return revision


def get_revision(revision_id):
    """ Take an id and return a reference to the revision """
    try:
        return Revision.objects.get(id=revision_id)
    except Revision.DoesNotExist:
        return None
