"""
This file demonstrates writing tests using the unittest module. These will pass
when you run "manage.py test".

Replace this with more appropriate tests for your application.
"""

from django.test import TestCase
from panic.models import Revision
from panic import store


class TestStoreAPI(TestCase):

    def test_get_view(self):
        """
        test by view id
        """
        view = store.create_view('test view', None)
        view = store.get_view(1)
        self.assertTrue(view is not None)

    def test_add_view(self):
        """
        add a new view with a name and get a view back, takes a parent id
        to reference
        """
        parent_view = store.create_view('Parent view', None)
        view = store.create_view('New view', parent_view)
        self.assertEquals(view.name, 'New view')
        self.assertEquals(view.parent.view_id, parent_view.view_id)

    def test_get_revision(self):
        """
        test retrieving a revision by id
        """
        store.add_revision(store.Revision(url='/'))
        rev = store.get_revision(1)
        self.assertEquals(1, rev.revision_id)
        self.assertEquals('/', rev.url)
        self.assertEquals(None, rev.parent)

    def test_add_revision(self):
        """
        test saving a revision to the storage API and getting a revision id
        """
        rev = store.ContentRevision(url='/', content='test content')
        self.assertEquals(None, rev.revision_id)
        store.add_revision(rev)
        self.assertEquals(1, rev.revision_id)

    def test_create_revision(self):
        """
        test creating a revision through the storage API
        """
        rev = store.create_revision(
            store.ContentRevision,
            url='/',
            content='test content')
        self.assertEquals(1, rev.revision_id)

    def test_create_child_revision(self):
        """
        test creating a child revision through the storage API
        """
        parent = store.create_revision(
            store.ContentRevision,
            url='/',
            content='parent content')
        self.assertEquals(1, parent.revision_id)

        child = store.create_revision(
            store.ContentRevision,
            url='/',
            content='child content',
            parent=parent)
        self.assertEquals(2, child.revision_id)
        self.assertEquals(parent.revision_id, child.parent.revision_id)


class TestViewAPI(TestCase):

    def test_save_page(self):
        """
        Tests setting a revision with a specific url to a view
        """
        view = store.create_view('test view', None)
        page = view.update_page(store.create_revision(store.Revision, url='/'))
        self.assertTrue(page is not None)
        self.assertEquals('/', page.url)
        self.assertEquals(1, page.revision_id)

    def test_get_page(self):
        """
        Tests retrieving a page from a view by it's url
        """
        view = store.create_view('test view', None)
        view.update_page(store.create_revision(store.Revision, url='/'))
        page = view.get_page('/')
        self.assertTrue(page is not None)
        self.assertEquals('/', page.url)

    def test_get_newer_revision(self):
        """
        Tests that when a new revision is set on a view, where there was an
        existing revision for the same url, the newer revision is retrieved for
        that url
        """
        view = store.create_view('test view', None)
        # First revision
        view.update_page(store.create_revision(store.Revision, url='/'))
        page = view.get_page('/')
        self.assertTrue(page is not None)
        self.assertEquals('/', page.url)
        self.assertEquals(1, page.revision_id)

        # Newer revision
        view.set_page(store.create_revision(
            store.Revision,
            url='/',
            parent=page))
        page = view.get_page('/')
        self.assertTrue(page is not None)
        self.assertEquals('/', page.url)
        self.assertEquals(2, page.revision_id)
        self.assertEquals(1, page.parent.revision_id)

    def test_get_content(self):
        """
        Test that ContentRevision is returned for a url where the
        Revision of that type
        """
        view = store.create_view('test view', None)
        view.update_page(store.create_revision(
            store.ContentRevision,
            url='/',
            content='test content'))
        page = view.get_page('/')
        self.assertTrue(page is not None)
        self.assertEquals('/', page.url)
        self.assertTrue(isinstance(page, store.ContentRevision))
        self.assertEquals('test content', page.content)

    def test_get_redirect(self):
        """
        Test that RedirectRevision is returned for a url where the
        Revision of that type
        """
        view = store.create_view('test view', None)
        view.update_page(store.create_revision(
            store.RedirectRevision,
            url='/',
            redirect_url='/test/'))
        page = view.get_page('/')
        self.assertTrue(page is not None)
        self.assertEquals('/', page.url)
        self.assertTrue(isinstance(page, store.RedirectRevision))
        self.assertEquals('/test/', page.redirect_url)

    def test_add_page(self):
        """
        Test creating a page for a url that did not have a page before
        """
        view = store.create_view('test view', None)

        try:
            page = view.get_page('/')
            self.fail('Page found for /, but it should not exist')
        except store.NoPageForUrl:
            pass

        view.update_page(store.create_revision(store.Revision, url='/'))
        page = view.get_page('/')
        self.assertTrue(page is not None)
        self.assertEquals('/', page.url)

    def test_move_page(self):
        """
        Test that moving a page puts a cloned revision at the new url, and
        places a redirect at the old url
        """
        view = store.create_view('test view', None)
        old_page = view.update_page(store.create_revision(
            store.ContentRevision,
            url='/',
            content='test content'))
        view.move_page(old_page, '/test/')

        # check redirect page
        redirect_page = view.get_page('/')
        self.assertEquals('/', redirect_page.url)
        self.assertTrue(isinstance(redirect_page, store.RedirectRevision))
        self.assertEquals('/test/', redirect_page.redirect_url)

        # check new page
        new_page = view.get_page(redirect_page.redirect_url)
        self.assertEquals('/test/', new_page.url)
        self.assertTrue(isinstance(new_page, store.ContentRevision))
        self.assertEquals('test content', new_page.content)
        self.assertEquals(new_page.parent.revision_id, old_page.revision_id)

    def test_branch_view(self):
        """
        Test creating a branch form an existing view
        """
        parent_view = store.create_view('parent view', None)
        page = parent_view.update_page(store.create_revision(
            store.ContentRevision,
            url='/',
            content='test content'))
        
        child_view = parent_view.branch('child view')
        self.assertEquals('child view', child_view.name)
        self.assertEquals(parent_view.view_id, child_view.parent.view_id)

        # Check that all the pages are the same
        self.assertEquals(parent_view.pages.count(), child_view.pages.count())
        for rev in parent_view.pages.all():
            self.assertTrue(child_view.pages.filter(pk=rev.revision_id).count() == 1)

    def test_diff_additions(self):
        parent_view = store.create_view('parent view', None)
        
        child_view = parent_view.branch('child view')
        new_page = child_view.update_page(store.create_revision(
            store.ContentRevision,
            url='/',
            content='new content'))

        (additions, changes, deletions) = child_view.diff(parent_view)
        self.assertEquals(1, len(additions))
        self.assertEquals(0, len(changes))
        self.assertEquals(0, len(deletions))

        self.assertEquals(additions['/'].revision_id, new_page.revision_id)
        self.assertEquals(additions['/'].content, 'new content')

    def test_diff_changes(self):
        parent_view = store.create_view('parent view', None)
        page = parent_view.update_page(store.create_revision(
            store.ContentRevision,
            url='/',
            content='test content'))
        
        child_view = parent_view.branch('child view')
        new_page = page.clone()
        new_page.content = 'new content'
        store.add_revision(new_page)
        child_view.update_page(new_page)

        (additions, changes, deletions) = child_view.diff(parent_view)
        self.assertEquals(0, len(additions))
        self.assertEquals(1, len(changes))
        self.assertEquals(0, len(deletions))

        self.assertTrue(changes['/'][0].revision_id < changes['/'][1].revision_id)
        
        self.assertEquals(changes['/'][0].revision_id, page.revision_id)
        self.assertEquals(changes['/'][0].content, 'test content')

        self.assertEquals(changes['/'][1].revision_id, new_page.revision_id)
        self.assertEquals(changes['/'][1].content, 'new content')

    def test_diff_deletions(self):
        parent_view = store.create_view('parent view', None)
        page = parent_view.update_page(store.create_revision(
            store.ContentRevision,
            url='/',
            content='test content'))
        
        child_view = parent_view.branch('child view')
        child_view.pages.remove(page)

        (additions, changes, deletions) = child_view.diff(parent_view)
        self.assertEquals(0, len(additions))
        self.assertEquals(0, len(changes))
        self.assertEquals(1, len(deletions))

        self.assertEquals(deletions['/'].revision_id, page.revision_id)
        self.assertEquals(deletions['/'].content, 'test content')

    def test_merge_view(self):
        pass


class TestRevisionAPI(TestCase):

    def test_content_revision_type(self):
        page = store.ContentRevision('/', '')
        self.assertEquals(page.revision_type, 'content')

    def test_redirect_revision_type(self):
        page = store.RedirectRevision('/', '/test')
        self.assertEquals(page.revision_type, 'redirect')

    def test_clone_content_revision(self):
        page = store.create_revision(store.ContentRevision, url='/', content='test content')
        newpage = page.clone()
        
        # check that data fields were copied to the new revision
        self.assertEquals(newpage.revision_type, 'content')
        self.assertEquals(newpage.url, '/')
        self.assertEquals(newpage.content, 'test content')
        self.assertEquals(newpage.parent.revision_id, page.revision_id)

        # new revision has no id until it has been added to the database
        self.assertEquals(newpage.revision_id, None)

    def test_clone_redirect_revision(self):
        page = store.create_revision(store.RedirectRevision, url='/', redirect_url='/test')
        newpage = page.clone()
        
        # check that data fields were copied to the new revision
        self.assertEquals(newpage.revision_type, 'redirect')
        self.assertEquals(newpage.url, '/')
        self.assertEquals(newpage.redirect_url, '/test')
        self.assertEquals(newpage.parent.revision_id, page.revision_id)

        # new revision has no id until it has been added to the database
        self.assertEquals(newpage.revision_id, None)


class TestGetPages(TestCase):

    def test_get_page(self):
        view = store.create_view('Test', None)
        self.client.session['current_view'] = view.view_id

        rev = store.create_revision(
            store.ContentRevision,
            url='/',
            content='test content')
        view.update_page(rev)
        
        response = self.client.get('/')
        self.assertEquals(response.status_code, 200)
        self.assertContains(response, 'test content')

    def test_get_latest(self):
        view = store.create_view('Test', None)
        self.client.session['current_view'] = view.view_id
        
        # set old content
        rev = store.create_revision(
            store.ContentRevision,
            url='/',
            content='old content')
        view.update_page(rev)

        # set new content
        rev = store.create_revision(
            store.ContentRevision,
            url='/',
            content='new content')
        view.update_page(rev)

        # check that we get the new content
        response = self.client.get('/')
        self.assertEquals(response.status_code, 200)
        self.assertContains(response, 'new content')
