from django.shortcuts import render_to_response, redirect
from django.http import Http404
from django.template import RequestContext

from panic import models, forms


def clean_path(func):
    """
    Make sure that the given path conforms to the standard path format.
    This means it starts with a /, and does not end with one.
    """
    def inner(request, path, *args, **kwds):
        path = '/' + path
        if len(path) > 2 and path[-1] == '/':
            path = path[:-1]
        return func(request, path, *args, **kwds)
    inner.__name__ = func.__name__
    return inner


@clean_path
def get_page(request, path):
    """
    Display the revision for the given path for the current site view.

    If the revision is a RedirectRevision, redirect the user to the target path
    """

    view = models.get_current_view(request)

    try:
        page = view.get_page(path)
    except models.NoPageForUrl:
        page = models.ContentRevision(url=path, content='Page not Found')

    if isinstance(page, models.RedirectRevision):
        return redirect(page.redirect_url)
        
    context = {
        'path': path,
        'page': page,
        'current_view': view,
        'available_views': models.SiteView.objects.all(),
        'add_page_form': forms.ContentRevisionForm(),
        'edit_page_form': forms.ContentRevisionForm(instance=page),
        'branch_form': forms.BranchForm(parent=view),
        'add_redirect_form': forms.RedirectRevisionForm(),
    }
    return render_to_response('home.html', context, RequestContext(request))

def get_revision(request, rev):
    view = models.get_current_view(request)

    try:
        page = models.ContentRevision.objects.get(pk=rev)
    except models.ContentRevision.DoesNotExist:
        page = models.ContentRevision(url='/', content='Revision not found')

    context = {
        'path': page.url,
        'page': page,
        'current_view': view,
        'available_views': models.SiteView.objects.all(),
        'add_page_form': forms.ContentRevisionForm(),
        'edit_page_form': forms.ContentRevisionForm(instance=page),
        'branch_form': forms.BranchForm(parent=view),
        'add_redirect_form': forms.RedirectRevisionForm(),
    }
    return render_to_response('revview.html', context, RequestContext(request))

def add_page(request):
    """
    Create a new ContentRevision in the current view
    """
    view = models.get_current_view(request)

    if request.method == 'POST':
        add_page_form = forms.ContentRevisionForm(request.POST)
        if add_page_form.is_valid():
            new_rev = add_page_form.save()
            view.update_page(new_rev)
            return redirect(new_rev.url)
    return redirect('/')

def add_redirect(request):
    """
    Create a new RedirectRevision in the current view
    """
    view = models.get_current_view(request)

    if request.method == 'POST':
        add_redirect_form = forms.RedirectRevisionForm(request.POST)
        if add_redirect_form.is_valid():
            new_rev = add_redirect_form.save()
            view.update_page(new_rev)
            return redirect(new_rev.redirect_url)
    return redirect('/')

@clean_path
def edit_page(request, path):
    """
    Create a cloned ContentRevision for the given page in the current view
    """
    view = models.get_current_view(request)

    try:
        page = view.get_page(path)
    except models.NoPageForUrl:
        page = models.ContentRevision(url=path, content='Page not Found')

    if request.method == 'POST':
        new_rev = page.clone()
        edit_page_form = forms.ContentRevisionForm(request.POST, instance=new_rev)
        if edit_page_form.is_valid():
            edit_page_form.save()
            view.update_page(new_rev)
            return redirect(new_rev.url)
    return redirect(path)

@clean_path
def switch_view(request, path, view_id):
    """
    Switch the viewing and editing context to a different site view
    """
    try:
        new_view = models.SiteView.objects.get(pk=view_id)
        models.set_current_view(request, new_view)
        return redirect(path)
    except models.SiteView.DoesNotExist:
        return redirect(path)

@clean_path
def branch_view(request, path):
    """
    Create a new site view from a given parent, copying all the current pages
    from the parent view
    """
    if request.method == 'POST':
        try:
            parent_view = models.get_view(request.POST['parent'])
        except:
            parent_view = models.get_current_view(request)
            
        new_view = parent_view.branch(request.POST['name'])
        models.set_current_view(request, new_view)
    return redirect(path)

def publish_view(request):
    """
    Merges that changes from the current site view into another site view
    """

    view = models.get_current_view(request)

    if 'target_view' in request.POST and request.POST['target_view'] != '':
        target_view = models.get_view(request.POST['target_view'])
    elif view.parent is not None:
        target_view = view.parent
    else:
        target_view = None

    if target_view is None:
        (additions, changes, deletions) = (None, None, None)
    else:
        if request.POST.get('publish_action', None) == 'publish':
            # do publishing
            target_view.update(view)
            models.set_current_view(request, target_view)
            return redirect('/')
            
        (additions, changes, deletions) = view.diff(target_view)

    context = {
        'path': '',
        'page': None,
        'current_view': view,
        'target_view': target_view,
        'additions': additions,
        'changes': changes,
        'deletions': deletions,
        'available_views': models.SiteView.objects.all(),
        'add_page_form': forms.ContentRevisionForm(),
        'edit_page_form': forms.ContentRevisionForm(),
        'branch_form': forms.BranchForm(parent=view),
        'add_redirect_form': forms.RedirectRevisionForm(),
    }
    return render_to_response('publish.html', context, RequestContext(request))
